from __future__ import annotations

import dataclasses

from numbers import Number

import numpy as np

from jijmodeling import Problem

from jijzept.entity.schema import SolverType
from jijzept.response import JijModelingResponse
from jijzept.sampler.base_sampler import (
    JijZeptBaseSampler,
    ParameterSearchParameters,
    merge_params_and_kwargs,
    sample_model,
)
from jijzept.sampler.sample_qubo import sample_qubo


@dataclasses.dataclass
class JijAmazonBraketDWaveParameters:
    num_reads: int | None = None


class JijAmazonBraketDWaveSampler(JijZeptBaseSampler):
    solver_type = SolverType(queue_name="amazonbraket_dwavesolver", solver="DWave")
    jijmodeling_solver_type = SolverType(
        queue_name="amazonbraket_dwavesolver", solver="DWaveParaSearch"
    )

    def sample_qubo(
        self,
        qubo: dict,
        parameters: JijAmazonBraketDWaveParameters | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        param_dict = merge_params_and_kwargs(
            parameters, kwargs, JijAmazonBraketDWaveParameters
        )
        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_qubo(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            qubo=qubo,
            constant=0,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )

        return sample_set

    def sample_model(
        self,
        problem: Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        multipliers: dict[str, Number] = {},
        fixed_variables: dict[str, dict[tuple[int, ...], int | float]] = {},
        search: bool = False,
        parameters: JijAmazonBraketDWaveParameters | None = None,
        algorithm: str | None = None,
        num_search: int | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using D-Wave via Amazon Braket.

        To configure the solver, instantiate the `JijAmazonBraketDWaveParameters` class and pass the instance to the `parameters` argument.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the
            placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from
            constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better
            values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option
            works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1000 will be set.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronization mode.
            queue_name (str, optional): queue_name.

        Returns:
           JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            import jijmodeling as jm
            n = jm.Placeholder('n')
            x = jm.Binary('x', shape=n)
            d = jm.Placeholder('d', shape=n)
            i = jm.Element("i", n)
            problem = jm.Problem('problem')
            problem += jm.Sum(i, d[i] * x[i])
            sampler = jz.JijAmazonBraketDWaveSampler(config='config.toml')
            response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
            ```
        """

        param_dict = merge_params_and_kwargs(
            parameters, kwargs, JijAmazonBraketDWaveParameters
        )

        para_search_params = ParameterSearchParameters(
            multipliers=multipliers,
            mul_search=search,
            num_search=num_search,
            algorithm=algorithm,
        )

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=problem,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set
