from __future__ import annotations

import itertools

import jijmodeling as jm
import numpy as np

from jijzept.client import JijZeptClient
from jijzept.sampler.base_sampler import ParameterSearchParameters, sample_model


def qubo_problem(
    qubo: dict[tuple[int, int], float] | dict[tuple[str, str], float], constant: float
) -> tuple[
    jm.Problem, dict[str, np.ndarray | list], dict[int, str], dict[str, jm.Binary]
]:
    qubo_keys = list(qubo.keys())
    unique_labels: dict[str, int] = {}
    N = jm.Placeholder("N")
    key = qubo_keys[0][0]
    if isinstance(key, int):
        # decision_variables = [jm.Binary("x", shape=N)]
        pass
    elif isinstance(key, str):
        unique_labels = {
            k: i for i, k in enumerate(itertools.chain.from_iterable(qubo_keys))
        }
        qubo_keys = [[unique_labels[e0], unique_labels[e1]] for e0, e1 in qubo_keys]
    else:
        TypeError(
            f"`{key.__class__.__name__}` object is not supported as a key of QUBO."
        )

    # Create QUBO Problem
    L = jm.Placeholder("L")
    E = jm.Placeholder("E", shape=(L, 2))
    Q = jm.Placeholder("Q", shape=L)
    c = jm.Placeholder("c")
    e = jm.Element("e", (0, L))
    x = jm.Binary("x", shape=N)
    qubo_model = jm.Problem("qubo")
    qubo_model += jm.Sum(e, Q[e] * x[E[e, 0]] * x[E[e, 1]]) + c

    edges = np.array(qubo_keys, dtype=int)
    values = np.array(list(qubo.values()), dtype=float)
    instance_data = {
        "N": int(np.max(edges)),
        "L": int(values.shape[0]),
        "E": edges.tolist(),
        "Q": values.tolist(),
        "c": constant,
    }

    return qubo_model, instance_data, {i: k for k, i in unique_labels.items()}, {"x": x}


def sample_qubo(
    client: JijZeptClient,
    solver: str,
    queue_name: str,
    qubo: dict[tuple[int, int], float],
    constant: float,
    max_wait_time: int | float | None,
    sync: bool,
    **kwargs,
):
    qubo_model, instance_data, var_map, _ = qubo_problem(qubo, constant)

    parameters = ParameterSearchParameters(
        multipliers={}, mul_search=False, normalize_qubo=False
    )

    sample_set = sample_model(
        client=client,
        solver=solver,
        queue_name=queue_name,
        problem=qubo_model,
        instance_data=instance_data,
        fixed_variables={},
        parameter_search_parameters=parameters,
        max_wait_time=max_wait_time,
        sync=sync,
        **kwargs,
    )
    sample_set.set_variable_labels(var_map)
    return sample_set
