from __future__ import annotations

from dataclasses import dataclass
from numbers import Number

import numpy as np

from jijmodeling import Problem

from jijzept.entity.schema import SolverType
from jijzept.response.jmresponse import JijModelingResponse
from jijzept.sampler.base_sampler import (
    JijZeptBaseSampler,
    ParameterSearchParameters,
    merge_params_and_kwargs,
    sample_model,
)


@dataclass
class JijDA3SolverParameters:
    """Manage Parameters for using Digital Annealer v3.

    Attributes:
        time_limit_sec (int): Set the timeout in seconds in the range 1 ~ 1800.
        target_energy (Optional[float]): Set the target energy value. The calculation is terminated when the minimum
        energy savings reaches the target energy.
        num_run (int): Set the number of parallel trial iterations in the range 1 ~ 16.
        num_group (int): Set the number of groups of parallel trials in the range 1 ~ 16.
        num_output_solution (int): Set the number of output solutions for each parallel trial group in the range 1 ~
        1024.
        gs_level (int): Set the level of global search. The higher this value, the longer the constraint utilization
        search will search in the range 0 ~ 100. If set the 1way 1hot or 2way 1hot, it is recommended that 0 be set
        for gs_level.
        gs_cutoff (int): Set the convergence decision level in the constraint utilization search of the solver in the
        range 0 ~ 1000000. If `0`, convergence judgement is off.
        penalty_auto_mode (int): Set the coefficient adjustment mode for the constaint term. If `0`, fixed to the
        value setin `penlaty_coef`. If `1`, the value set in `penalty_coef` is used as the initial value and adjusted
        automatically.
        penalty_coef (int): Set the coefficients of the constraint term.
        penalty_inc_rate (int): Set parameters for automatic adjustment of constriant term.
        max_penalty_coef (int): Set the maximum value of the constraint term coefficient in the global search. If no
        maximum value is specified, set to 0.
    """

    time_limit_sec: int = 10
    target_energy: float | None = None
    num_run: int = 16
    num_group: int = 1
    num_output_solution: int = 5
    gs_level: int = 5
    gs_cutoff: int = 8000
    penalty_auto_mode: int = 1
    penalty_coef: int = 1
    penalty_inc_rate: int = 150
    max_penalty_coef: int = 0


class JijDA3Sampler(JijZeptBaseSampler):
    jijmodeling_solver_type: SolverType = SolverType(
        queue_name="thirdpartysolver",
        solver="DigitalAnnealer",
    )

    def __init__(
        self,
        token: str | None = None,
        url: str | dict = None,
        proxy: str | None = None,
        config: str | None = None,
        config_env: str = "default",
        da3_token: str = "",
        da3_url: str = "https://api.aispf.global.fujitsu.com/da",
    ) -> None:
        """Sets Jijzept token and url and Digital Annealer v3 token and url.

        Args:
            token (Optional[str]): Token string.
            url (Optional[Union[str, dict]]): API URL.
            proxy (Optional[str]): Proxy URL.
            config (Optional[str]): Config file path for JijZept.
            config_env (str): config env.
            da3_token (str): Token string for Degital Annealer 3.
            da3_url (Optional[str]): API Url string for Degital Annealer 3.
        """
        super().__init__(token, url, proxy, config, config_env)

        self.da3_token = da3_token
        self.da3_url = da3_url

    def sample_model(
        self,
        model: Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        fixed_variables: dict[str, dict[tuple[int, ...], Number]] = {},
        inequalities_lambda: dict[str, int] = {},
        parameters: JijDA3SolverParameters | None = None,
        normalize_qubo: bool = False,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using JijModeling by means of Digital Annealer v3.

        To configure the solver, instantiate the `JijDA3SolverParameters` class and pass the instance to the `parameters` argument.

        Args:
            model (Problem): Mathematical expression of JijModeling.
            feed_dict (dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the
            placeholders.
            fixed_variables (dict[str, dict[tuple[int, ...], Number]]): Dictionary of variables to fix.
            inequalities_lambda (dict[str, int]): Coefficient of inequality. If omitted, set to 1. The coefficients
            of the equality constraints can be set from JijDA3SolverParameters.
            parameters (Optional[JijDA3SolverParameters]): Parameters used in Ditital Annealer 3. If `None`,
            the default value of the JijDA3SolverParameters will be set.
            normalize_qubo (bool): Option to normalize the QUBO coefficient. Defaults to `False`.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool): Synchronous mode.
            queue_name (Optional[str]): Queue name.
            kwargs: Digital Annealer 3 parameters using **kwags. If both `**kwargs` and `parameters` are exist,
            the value of `**kwargs` takes precedence.

        Returns:
            JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijmodeling as jm
            from jijzept import JijDA3Sampler, JijDA3SolverParameters

            w = jm.Placeholder("w", dim=1)
            num_items = jm.Placeholder("num_items")
            c = jm.Placeholder("c")
            y = jm.Binary("y", shape=(num_items,))
            x = jm.Binary("x", shape=(num_items, num_items))
            i = jm.Element("i", num_items)
            j = jm.Element("j", num_items)
            problem = jm.Problem("bin_packing")
            problem += y[:]
            problem += jm.Constraint("onehot_constraint", jm.Sum(j, x[i, j]) - 1 == 0, forall=i)
            problem += jm.Constraint(
                "knapsack_constraint", jm.Sum(i, w[i] * x[i, j]) - y[j] * c <= 0, forall=j
            )

            feed_dict = {"num_items": 2, "w": [9, 1], "c": 10}
            inequalities_lambda = {"knapsack_constraint": 22}

            sampler = JijDA3Sampler(config="xx", da3_token="oo")
            parameters = JijDA3SolverParameters(penalty_coef=2)

            sampleset = sampler.sample_model(
                problem, feed_dict, inequalities_lambda=inequalities_lambda, parameters=parameters
            )
            ```
        """

        param_dict = merge_params_and_kwargs(parameters, kwargs, JijDA3SolverParameters)
        para_search_params = ParameterSearchParameters(
            multipliers={},
            mul_search=False,
        )

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        param_dict["token"] = self.da3_token
        param_dict["url"] = self.da3_url
        param_dict["normalize_qubo"] = normalize_qubo
        param_dict["inequalities_lambda"] = inequalities_lambda

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=model,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set


@dataclass
class JijDA4SolverParameters:
    """Manage Parameters for using Digital Annealer v4.

    Attributes:
        time_limit_sec (int): Set the timeout in seconds in the range 1 ~ 1800.
        target_energy (Optional[float]): Set the target energy value. The calculation is terminated when the minimum
        energy savings reaches the target energy.
        num_run (int): Set the number of parallel trial iterations in the range 1 ~ 16.
        num_group (int): Set the number of groups of parallel trials in the range 1 ~ 16.
        num_output_solution (int): Set the number of output solutions for each parallel trial group in the range 1 ~
        1024.
        gs_level (int): Set the level of global search. The higher this value, the longer the constraint utilization
        search will search in the range 0 ~ 100. If set the 1way 1hot or 2way 1hot, it is recommended that 0 be set
        for gs_level.
        gs_cutoff (int): Set the convergence decision level in the constraint utilization search of the solver in the
        range 0 ~ 1000000. If `0`, convergence judgement is off.
        one_hot_level (int): Levels of 1hot constraints search.
        one_hot_cutoff (int): Convergence decision level for 1hot constraints search. If 0 is set, this function is
        turned off.
        internal_penalty (int): 1hot constraint internal generation mode. Note that if 1way- or 2way 1hot constraints
        are specified to `jijmodeling.Constraint`, internal_penalty is set to 1 internally, regardless of user input.
        penalty_auto_mode (int): Set the coefficient adjustment mode for the constaint term. If `0`, fixed to the
        value setin `penlaty_coef`. If `1`, the value set in `penalty_coef` is used as the initial value and adjusted
        automatically.
        penalty_coef (int): Set the coefficients of the constraint term.
        penalty_inc_rate (int): Set parameters for automatic adjustment of constriant term.
        max_penalty_coef (int): Set the maximum value of the constraint term coefficient in the global search. If no
        maximum value is specified, set to 0.
    """

    time_limit_sec: int = 10
    target_energy: float | None = None
    num_run: int = 16
    num_group: int = 1
    num_output_solution: int = 5
    gs_level: int = 5
    gs_cutoff: int = 8000
    one_hot_level: int = 3
    one_hot_cutoff: int = 100
    internal_penalty: int = 0
    penalty_auto_mode: int = 1
    penalty_coef: int = 1
    penalty_inc_rate: int = 150
    max_penalty_coef: int = 0


class JijDA4Sampler(JijZeptBaseSampler):
    jijmodeling_solver_type: SolverType = SolverType(
        queue_name="da4solver",
        solver="DigitalAnnealerV4",
    )

    def __init__(
        self,
        token: str | None = None,
        url: str | None = None,
        proxy: str | None = None,
        config: str | None = None,
        config_env: str = "default",
        da4_token: str = "",
        da4_url: str = "https://api.aispf.global.fujitsu.com/da",
    ) -> None:
        """Sets Jijzept token and url and Digital Annealer v4 token and url.

        Args:
            token (Optional[str]): Token string.
            url (Optional[Union[str, dict]]): API URL.
            proxy (Optional[str]): Proxy URL.
            config (Optional[str]): Config file path for JijZept.
            config_env (str): config env.
            da4_token (str): Token string for Degital Annealer 4.
            da4_url (Optional[str]): API Url string for Degital Annealer 4.
        """
        super().__init__(
            token=token, url=url, proxy=proxy, config=config, config_env=config_env
        )
        self.da4_token = da4_token
        self.da4_url = da4_url

    def sample_model(
        self,
        model: Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        fixed_variables: dict[str, dict[tuple[int, ...], Number]] = {},
        inequalities_lambda: dict[str, int] = {},
        parameters: JijDA4SolverParameters | None = None,
        normalize_qubo: bool = False,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using JijModeling by means of Digital Annealer v4.

        Note that this sampler solve problems with using Azure Blob Storage, which is a feature of Digital Annealer
        v4, and there is no option not to use Azure Blob Storage.

        To configure the solver, instantiate the `JijDA4SolverParameters` class and pass the instance to the `parameters` argument.

        Args:
            model (Problem): Mathematical expression of JijModeling.
            feed_dict (dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the
            placeholders.
            fixed_variables (dict[str, dict[tuple[int, ...], Number]]): Dictionary of variables to fix.
            inequalities_lambda (dict[str, int]): Coefficient of inequality. If omitted, set to 1. The coefficients
            of the equality constraints can be set from JijDA4SolverParameters.
            parameters (Optional[JijDA4SolverParameters]): Parameters used in Ditital Annealer 4. If `None`,
            the default value of the JijDA4SolverParameters will be set.
            normalize_qubo (bool): Option to normalize the QUBO coefficient and inequality constraint conditions.
            Defaults to `False`.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool): Synchronous mode.
            queue_name (Optional[str]): Queue name.
            kwargs: Digital Annealer 4 parameters using **kwags. If both `**kwargs` and `parameters` are exist,
            the value of `**kwargs` takes precedence.

        Returns:
            JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijmodeling as jm
            from jijzept import JijDA4Sampler, JijDA4SolverParameters

            w = jm.Placeholder("w", dim=1)
            num_items = jm.Placeholder("num_items")
            c = jm.Placeholder("c")
            y = jm.Binary("y", shape=(num_items,))
            x = jm.Binary("x", shape=(num_items, num_items))
            i = jm.Element("i", num_items)
            j = jm.Element("j", num_items)
            problem = jm.Problem("bin_packing")
            problem += y[:]
            problem += jm.Constraint("onehot_constraint", jm.Sum(j, x[i, j]) - 1 == 0, forall=i)
            problem += jm.Constraint(
                "knapsack_constraint", jm.Sum(i, w[i] * x[i, j]) - y[j] * c <= 0, forall=j
            )

            feed_dict = {"num_items": 2, "w": [9, 1], "c": 10}
            inequalities_lambda = {"knapsack_constraint": 22}

            sampler = JijDA4Sampler(config="xx", da4_token="oo")
            parameters = JijDA4SolverParameters(penalty_coef=2)

            sampleset = sampler.sample_model(
                problem, feed_dict, inequalities_lambda=inequalities_lambda, parameters=parameters
            )
            ```
        """

        param_dict = merge_params_and_kwargs(parameters, kwargs, JijDA4SolverParameters)

        para_search_params = ParameterSearchParameters(
            multipliers={},
            mul_search=False,
        )

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        param_dict["token"] = self.da4_token
        param_dict["url"] = self.da4_url
        param_dict["normalize_qubo"] = normalize_qubo
        param_dict["inqualities_lambda"] = inequalities_lambda

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=model,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set
